#ifdef HAMMER_LLVM_BACKEND
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wpedantic"
#include <llvm-c/Core.h>
#pragma GCC diagnostic pop
#include "../backends/llvm/llvm.h"
#endif
#include "parser_internal.h"

static HParseResult* parse_nothing(void* x,HParseState* y) {
  (void)(x);
  (void)(y);
  // not a mistake, this parser always fails
  return NULL;
}

static void desugar_nothing(HAllocator *mm__, HCFStack *stk__, void *env) {
  HCFS_BEGIN_CHOICE() {
  } HCFS_END_CHOICE();
}

static bool nothing_ctrvm(HRVMProg *prog, void* env) {
  h_rvm_insert_insn(prog, RVM_MATCH, 0x0000);
  h_rvm_insert_insn(prog, RVM_MATCH, 0xFFFF);
  return true;
}

#ifdef HAMMER_LLVM_BACKEND

static bool nothing_llvm(HLLVMParserCompileContext *ctxt,
                         HParser *p,
                         LLVMValueRef *res) {
  /* Allocator */
  HAllocator *mm__;
  /* Names */
  char *nothing_entry_bb_name = NULL;
  int nothing_entry_bb_name_len = 0;
  /* Nothing entry bb */
  LLVMBasicBlockRef entry;
  /* Nothing result */
  LLVMValueRef rv;

  if (!res) return false;
  if (!p) return false;
  if (!ctxt) return false;

  /* Unpack allocator */
  mm__ = ctxt->mm__;

  /* This one just always returns NULL */
  nothing_entry_bb_name_len = snprintf(NULL, 0, "nothing_entry_%p", p);
  nothing_entry_bb_name = h_new(char, nothing_entry_bb_name_len + 1);
  snprintf(nothing_entry_bb_name, nothing_entry_bb_name_len + 1,
           "nothing_entry_%p", p);
  entry = LLVMAppendBasicBlock(ctxt->func, nothing_entry_bb_name);
  LLVMBuildBr(ctxt->builder, entry);
  LLVMPositionBuilderAtEnd(ctxt->builder, entry);
  rv = LLVMConstNull(ctxt->llvm_parseresultptr);

  /* Pass it back out */
  *res = rv;

  if (nothing_entry_bb_name) {
    h_free(nothing_entry_bb_name);
    nothing_entry_bb_name = NULL;
  }

  return true;
}

#endif /* defined(HAMMER_LLVM_BACKEND) */

static const HParserVtable nothing_vt = {
  .parse = parse_nothing,
  .isValidRegular = h_true,
  .isValidCF = h_true,
  .desugar = desugar_nothing,
  .compile_to_rvm = nothing_ctrvm,
#ifdef HAMMER_LLVM_BACKEND
  .llvm = nothing_llvm,
#endif
  .higher = false,
};

HParser* h_nothing_p() {
  return h_nothing_p__m(&system_allocator);
}
HParser* h_nothing_p__m(HAllocator* mm__) { 
  return h_new_parser(mm__, &nothing_vt, NULL);
}
